from inspect import signature
import sys

EMPTY_SIGNATURE_TYPE = "<class 'inspect._empty'>"


"""
A python dectorater to enforce type checks on your python functions
for your stupid colleagues who pass int type when you clearly told them to
pass string.
ヽ(ಠ_ಠ) ノ

"""


def parse_arguments(signature_object):
    """
    A function to parse signature object to return argument type

    parameters:
    signature_object -> Signature

    returns:
    types_list -> List
    """
    return ([repr(p.annotation).ljust(13)
             for p in signature_object.parameters.values()], signature_object.return_annotation)


def match_type(types, *args):
    """
   Matches types defined in the fucntion name to the given arguments/

   Parameters:
   types -> list
   args -> list

   returns:
   None if arguments match defined types
   Raises TypeError if not
    """

    for x, y in zip(types, args):
        if(str(type(y)) != x and x != EMPTY_SIGNATURE_TYPE):
            raise TypeError(
                f"Type mismatch value of type {type(y)} can not be assigned to parameter of type {x}")


def match_return_type(sig, val):
    if(type(val) != sig and str(sig) != EMPTY_SIGNATURE_TYPE):
        raise TypeError(
            f"Type mismatch value of type {type(val)} can not be assigned to return of type {sig}")
    return val


def watcher(f):
    def inner(*args, **kwargs):
        types, return_type = parse_arguments(signature(f))
        try:
            return match_type(
                types, *args) or match_return_type(return_type, f(*args, **kwargs))
        except TypeError as ve:
            tb = sys.exc_info()[2]
            raise ve.with_traceback(tb)
    return inner



from watcher import watcher
import pytest


class User:
    pass


@watcher
def function_without_types(a, b) -> str:
    return 'pass'


@watcher
def function_with_types(a: str, b: list):
    return 'pass'


@watcher
def function_with_userdef_types(a: str, b: User):
    return "pass"


"""
Test cases for function without types
"""


def test_function_without_types_success():
    assert function_without_types("a", 1) == "pass"


def test_function_without_types_fail():
    with pytest.raises(TypeError):
        function_without_types()


"""
Test cases for functions with types
"""


def test_function_with_types_success():
    assert function_with_types("a", []) == "pass"


def test_function_with_types_fail_TypeError_first_arg():
    with pytest.raises(TypeError):
        function_with_types(1, [])


def test_function_with_types_fail_TypeError_second_arg():
    with pytest.raises(TypeError):
        function_with_types("a", "a")


def test_function_with_types_fail_TypeError_both_arg():
    with pytest.raises(TypeError):
        function_with_types(1, 1)


def test_function_with_types_fail_TypeError_extra_arg():
    with pytest.raises(TypeError):
        function_with_types("1", [], "0")


def test_function_with_types_fail_TypeError_no_arg():
    with pytest.raises(TypeError):
        function_with_types()


"""
Test for User defined Data Types
"""


def test_function_with_userdef_types_success():
    assert function_with_userdef_types("a", User()) == "pass"


def test_function_with_userdef_types_fail():
    with pytest.raises(TypeError):
        function_with_userdef_types("a", 1) == "pass"



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(name='be_holder',
      version='1.1',
      description='A python module for type checking function parameters and returns',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/wassafshahzad/beholder',
      author='Wassaf Shahzad',
      author_email='wassafshahzad@gmail.com',
      license='MIT',
      packages=['be_holder'],
      zip_safe=False)
